module Wiki (
	Backend (getCurrent,getList,get,setCurrent,update,listKeys)
	,PGB
	,createPGB
	,Markup (Text, Paragraph, Link, Bold, Emph, Heading, Url, Underline, Strike, Pre)
	,Document
	,wikiParser
	,htmlOutput

) where

import Dbconnect
import Data.Char
import Text.ParserCombinators.Parsec

class Backend a where
	--Keyword -> (Full text,date)
	getCurrent :: a -> String -> IO (Maybe (String,String))

	--Keyword -> [(id,date,author,comment)]
	getList :: a -> String -> IO [(String, String, String, String)]

	--Keyword -> id -> Full text
	get :: a -> String -> String -> IO (Maybe String)

	--Keyword -> id -> ()
	setCurrent :: a -> String -> String -> IO Bool

	--Keyword -> Full text -> id
	update :: a -> String -> String -> String -> String -> IO String

	--[Keyword]
	listKeys :: a -> IO [String]

data Markup = Text String
	| Paragraph 
	| Link String String
	| Bold [Markup]
	| Emph [Markup]
	| Heading Int [Markup]
	| Url String
	| Underline [Markup]
	| Strike [Markup]
	| Pre String

type Document = [Markup]

wikiParser :: GenParser Char st Document	
wikiParser = do 
	s <- pMain
	ss <- (wikiParser <|> return [])
	return (s:ss)

pPara = do
	pEol
	pEol
	return (Paragraph)


pSpace = do
	space 
	return (Text " ")

pEol = char '\n' <|> do
	char '\r'
	char '\n'

pLinkParser = do try (string "]]" >> return ([],[])) 
	<|> try (do
		string " | "
		d <- pString "]]"
		return ([],d))
	<|> (do
		c <- anyChar
		(l,d) <- pLinkParser
		return (c:l,d))

pLink = do
	string "[["
	(l,d) <- pLinkParser
	case d of
		[] -> return (Link l l)
		_ -> return (Link l d)

pBold = pS "**" (\s -> Bold s)
pEmph = pS "//" (\s -> Emph s)
pUnderline = pS "__" (\s -> Underline s)
pStrike = pS "--" (\s -> Strike s)

pS s f = do
	string s
	s <- pStopAt s
	return (f s)

pStopAt xs = do 
	try (string xs >> return []) <|> (do
		s <- pMain 
		ss <- pStopAt xs
		return (s:ss))

pString xs = do
	try (string xs >> return []) <|> (do
	s <- anyChar
	ss <- pString xs
	return (s:ss))
	
pHeading = do
	many1 pEol
	level <- many1 $ char '='
	char ' '
	s <- pStopAt (' ':level)
	return (Heading (length level) s)

pURL = do
	proto <- many1 letter
	string "://"
	s <- many1 (alphaNum <|> oneOf "?.:&-/")
	return (Url (proto++"://"++s))

pPre = do
	string "<pre>"
	s <- pString "</pre>"
	return (Pre s)
	
pMain = try (pHeading)
	<|> (try (pPara) 
	<|> pSpace 
	<|> try(pPre) 
	<|> try(pBold) 
	<|> try(pEmph) 
	<|> try(pUnderline) 
	<|> try(pStrike) 
	<|> try (pLink)
	<|> try (pURL)
	<|> pOtherChar
	<|> pText)	

pOtherChar = do
	c <- oneOf "*,;.:!?[]()'\"=-%$<>/\\|"
	return (Text (c:[]))

pText = do
	t <- many1 alphaNum--(noneOf ['\n','\r','[',']'])
	return (Text t)


newtype PGB = PGB DBService

createPGB :: String -> String -> String -> String -> IO PGB
createPGB host database user password = let db = createDBService host database "" user password Nothing in return (PGB db)


testDB = createPGB "wave" "wiki" "wiki" "12wiki34db"

instance Backend PGB where

	getCurrent (PGB db) key = do 
		result <- selectReturnTuples db $ "SELECT fulltext,timestamp FROM curtexts WHERE keyword="++tov key
		case result of
			[text,date]:_ -> return (Just (text,date))
			_ -> return Nothing

	getList (PGB db) key = do
		list <- selectReturnTuples db $ "SELECT id, timestamp, author, comment from fulltexts WHERE keyword = "++tov key
		case list of
			[[]] -> return []
			_ -> return $ map (\[id,date,author,comment] -> (id,date,author,comment)) list

	get (PGB db) key id = do
		list <- selectReturnTuples db $ "SELECT fulltext from fulltexts WHERE id = "++tov id	
		case list of
			[s]:_ -> return (Just s)
			_ -> return Nothing

	setCurrent (PGB db) key id = do
		full <- selectReturnTuples db $ "SELECT keyword FROM fulltexts WHERE keyword="++tov key++" AND id='"++id++"'"
		cur <- selectReturnTuples db $ "SELECT keyword FROM curtexts WHERE keyword="++tov key
		case full of
			[[]] -> do
				return False
			_  -> do
				rows <- case cur of
					[[]] -> do
						execute db $ "INSERT INTO current (keyword, id) VALUES ("++tov key++","++tov id++")"
					_  -> do
						execute db $ "UPDATE current SET id = "++tov id++" WHERE keyword = "++tov key
				if rows == 1 then return True
					else return False 

	update (PGB db) key text author comment = do
		rows <- execute db $ "INSERT INTO fulltexts (keyword,fulltext, author, comment) VALUES ("++tov key++","++tov text++","++tov author++","++tov comment++")"
		if rows == 0 then return ""
			else do
				[[id]] <- selectReturnTuples db $ "SELECT currval('fulltexts_id_seq')" 
				setCurrent (PGB db) key id
				return id

	listKeys (PGB db) = do
		list <- selectReturnTuples db $ "SELECT keyword FROM fulltexts GROUP BY keyword ORDER BY lower(keyword)"
		case list of
			[[]] -> return []
			_ -> mapM (\[key] -> return key) list
		
			

tov :: String -> String
tov s = '\'':escapeQuery s++"'"
		
toHtml :: [Markup] -> String
toHtml [] = []
toHtml ((Paragraph):xs) = "<p>\n"++toHtml xs
toHtml ((Text s):xs) = s++toHtml xs
toHtml ((Link l d):xs) = "<link: "++l++" desc: "++d++">"++toHtml xs
toHtml ((Bold d):xs) = "<b>"++toHtml d++"</b>"++toHtml xs
toHtml ((Emph d):xs) = "<em>"++toHtml d++"</em>"++toHtml xs
toHtml ((Underline d):xs) = "<u>"++toHtml d++"</u>"++toHtml xs
toHtml ((Strike d):xs) = "<strike>"++toHtml d++"</strike>"++toHtml xs
toHtml ((Heading n d):xs) = "\n<h"++show n++">"++toHtml d++"</h"++show n++">\n"++toHtml xs
toHtml ((Url l):xs) = "<link: "++l++">"++toHtml xs
toHtml ((Pre s):xs) = "<pre>"++s++"</pre>"++toHtml xs


htmlOutput s = case parse wikiParser "" s of
					Right n -> do 
						putStr (toHtml n)
					Left e -> do
						print e
